@@ -1,6 +1,6 @@
 /*
  * BlueALSA - bluealsa-dbus.c
- * Copyright (c) 2016-2023 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -9,7 +9,10 @@
  */
 
 #include "bluealsa-dbus.h"
-/* IWYU pragma: no_include "config.h" */
+
+#if HAVE_CONFIG_H
+# include <config.h>
+#endif
 
 #include <errno.h>
 #include <fcntl.h>
@@ -31,11 +34,11 @@
 
 #include "a2dp.h"
 #include "ba-adapter.h"
+#include "ba-config.h"
 #include "ba-device.h"
 #include "ba-transport.h"
-#include "bluealsa-config.h"
+#include "ba-transport-pcm.h"
 #include "bluealsa-iface.h"
-#include "bluealsa-skeleton.h"
 #include "bluez.h"
 #include "dbus.h"
 #include "hfp.h"
@@ -72,7 +75,7 @@ static GVariant *ba_variant_new_bluealsa_adapters(void) {
 
 static GVariant *ba_variant_new_bluealsa_profiles(void) {
 
-	struct {
+	const struct {
 		const char *name;
 		bool enabled;
 	} profiles[] = {
@@ -85,6 +88,9 @@ static GVariant *ba_variant_new_bluealsa_profiles(void) {
 		{ BLUEALSA_TRANSPORT_TYPE_HFP_HF, config.profile.hfp_hf },
 		{ BLUEALSA_TRANSPORT_TYPE_HSP_AG, config.profile.hsp_ag },
 		{ BLUEALSA_TRANSPORT_TYPE_HSP_HS, config.profile.hsp_hs },
+#if ENABLE_MIDI
+		{ BLUEALSA_TRANSPORT_TYPE_MIDI, config.profile.midi },
+#endif
 	};
 
 	const char *strv[ARRAYSIZE(profiles)];
@@ -133,13 +139,16 @@ static GVariant *ba_variant_new_bluealsa_codecs(void) {
 		BLUEALSA_TRANSPORT_TYPE_HFP_HF,
 	};
 
-	struct {
-		uint16_t codec_id;
+	const struct {
+		uint8_t codec_id;
 		bool enabled;
 	} hfp_codecs[] = {
 		{ HFP_CODEC_CVSD, config.hfp.codecs.cvsd },
 #if ENABLE_MSBC
 		{ HFP_CODEC_MSBC, config.hfp.codecs.msbc },
+#endif
+#if ENABLE_LC3_SWB
+		{ HFP_CODEC_LC3_SWB, config.hfp.codecs.lc3_swb },
 #endif
 	};
 
@@ -193,6 +202,10 @@ static GVariant *ba_variant_new_transport_type(const struct ba_transport *t) {
 		return g_variant_new_string(BLUEALSA_TRANSPORT_TYPE_HSP_AG);
 	case BA_TRANSPORT_PROFILE_HSP_HS:
 		return g_variant_new_string(BLUEALSA_TRANSPORT_TYPE_HSP_HS);
+#if ENABLE_MIDI
+	case BA_TRANSPORT_PROFILE_MIDI:
+		return g_variant_new_string(BLUEALSA_TRANSPORT_TYPE_MIDI);
+#endif
 	case BA_TRANSPORT_PROFILE_NONE:
 		break;
 	}
@@ -221,7 +234,7 @@ static GVariant *ba_variant_new_pcm_mode(const struct ba_transport_pcm *pcm) {
 }
 
 static GVariant *ba_variant_new_pcm_running(const struct ba_transport_pcm *pcm) {
-	return g_variant_new_boolean(pcm->th->state == BA_TRANSPORT_THREAD_STATE_RUNNING);
+	return g_variant_new_boolean(ba_transport_pcm_state_check_running(pcm));
 }
 
 static GVariant *ba_variant_new_pcm_format(const struct ba_transport_pcm *pcm) {
@@ -240,9 +253,9 @@ static GVariant *ba_variant_new_pcm_codec(const struct ba_transport_pcm *pcm) {
 	const struct ba_transport *t = pcm->t;
 	const char *codec = NULL;
 	if (t->profile & BA_TRANSPORT_PROFILE_MASK_A2DP)
-		codec = a2dp_codecs_codec_id_to_string(t->codec_id);
+		codec = a2dp_codecs_codec_id_to_string(ba_transport_get_codec(t));
 	if (t->profile & BA_TRANSPORT_PROFILE_MASK_SCO)
-		codec = hfp_codec_id_to_string(t->codec_id);
+		codec = hfp_codec_id_to_string(ba_transport_get_codec(t));
 	if (codec != NULL)
 		return g_variant_new_string(codec);
 	return NULL;
@@ -260,6 +273,10 @@ static GVariant *ba_variant_new_pcm_delay(const struct ba_transport_pcm *pcm) {
 	return g_variant_new_uint16(ba_transport_pcm_get_delay(pcm));
 }
 
+static GVariant *ba_variant_new_pcm_delay_adjustment(const struct ba_transport_pcm *pcm) {
+	return g_variant_new_int16(ba_transport_pcm_delay_adjustment_get(pcm));
+}
+
 static GVariant *ba_variant_new_pcm_soft_volume(const struct ba_transport_pcm *pcm) {
 	return g_variant_new_boolean(pcm->soft_volume);
 }
@@ -278,29 +295,6 @@ static GVariant *ba_variant_new_pcm_volume(const struct ba_transport_pcm *pcm) {
 	return g_variant_new_uint16((ch1 << 8) | (pcm->channels == 1 ? 0 : ch2));
 }
 
-static void ba_variant_populate_pcm(GVariantBuilder *props, const struct ba_transport_pcm *pcm) {
-
-	GVariant *value;
-	g_variant_builder_init(props, G_VARIANT_TYPE("a{sv}"));
-
-	g_variant_builder_add(props, "{sv}", "Device", ba_variant_new_device_path(pcm->t->d));
-	g_variant_builder_add(props, "{sv}", "Sequence", ba_variant_new_device_sequence(pcm->t->d));
-	g_variant_builder_add(props, "{sv}", "Transport", ba_variant_new_transport_type(pcm->t));
-	g_variant_builder_add(props, "{sv}", "Mode", ba_variant_new_pcm_mode(pcm));
-	g_variant_builder_add(props, "{sv}", "Running", ba_variant_new_pcm_running(pcm));
-	g_variant_builder_add(props, "{sv}", "Format", ba_variant_new_pcm_format(pcm));
-	g_variant_builder_add(props, "{sv}", "Channels", ba_variant_new_pcm_channels(pcm));
-	g_variant_builder_add(props, "{sv}", "Sampling", ba_variant_new_pcm_sampling(pcm));
-	if ((value = ba_variant_new_pcm_codec(pcm)) != NULL)
-		g_variant_builder_add(props, "{sv}", "Codec", value);
-	if ((value = ba_variant_new_pcm_codec_config(pcm)) != NULL)
-		g_variant_builder_add(props, "{sv}", "CodecConfiguration", value);
-	g_variant_builder_add(props, "{sv}", "Delay", ba_variant_new_pcm_delay(pcm));
-	g_variant_builder_add(props, "{sv}", "SoftVolume", ba_variant_new_pcm_soft_volume(pcm));
-	g_variant_builder_add(props, "{sv}", "Volume", ba_variant_new_pcm_volume(pcm));
-
-}
-
 static bool ba_variant_populate_sep(GVariantBuilder *props, const struct a2dp_sep *sep) {
 
 	const struct a2dp_codec *codec;
@@ -312,10 +306,10 @@ static bool ba_variant_populate_sep(GVariantBuilder *props, const struct a2dp_se
 	if (!codec->enabled)
 		return false;
 
-	uint8_t caps[sizeof(sep->capabilities)];
+	a2dp_t caps = sep->capabilities;
 	size_t size = MIN(sep->capabilities_size, sizeof(caps));
-	if (a2dp_filter_capabilities(codec, memcpy(caps, &sep->capabilities, size), size) != 0) {
-		error("Couldn't filter %s capabilities: %s",
+	if (a2dp_filter_capabilities(codec, &codec->capabilities, &caps, size) != 0) {
+		error("Couldn't filter %s SEP capabilities: %s",
 				a2dp_codecs_codec_id_to_string(sep->codec_id),
 				strerror(errno));
 		return false;
@@ -323,42 +317,7 @@ static bool ba_variant_populate_sep(GVariantBuilder *props, const struct a2dp_se
 
 	g_variant_builder_init(props, G_VARIANT_TYPE("a{sv}"));
 	g_variant_builder_add(props, "{sv}", "Capabilities", g_variant_new_fixed_array(
-				G_VARIANT_TYPE_BYTE, caps, size, sizeof(uint8_t)));
-
-	switch (codec->codec_id) {
-	case A2DP_CODEC_SBC:
-		break;
-#if ENABLE_MPEG
-	case A2DP_CODEC_MPEG12:
-		break;
-#endif
-#if ENABLE_AAC
-	case A2DP_CODEC_MPEG24:
-		break;
-#endif
-#if ENABLE_APTX
-	case A2DP_CODEC_VENDOR_APTX:
-		break;
-#endif
-#if ENABLE_APTX_HD
-	case A2DP_CODEC_VENDOR_APTX_HD:
-		break;
-#endif
-#if ENABLE_FASTSTREAM
-	case A2DP_CODEC_VENDOR_FASTSTREAM:
-		break;
-#endif
-#if ENABLE_LC3PLUS
-	case A2DP_CODEC_VENDOR_LC3PLUS:
-		break;
-#endif
-#if ENABLE_LDAC
-	case A2DP_CODEC_VENDOR_LDAC:
-		break;
-#endif
-	default:
-		g_assert_not_reached();
-	}
+				G_VARIANT_TYPE_BYTE, &caps, size, sizeof(uint8_t)));
 
 	return true;
 }
@@ -389,10 +348,10 @@ void bluealsa_dbus_register(void) {
 		.get_property = bluealsa_manager_get_property,
 	};
 
-	debug("Registering D-Bus manager: %s", bluealsa_dbus_manager_path);
+	debug("Registering BlueALSA D-Bus manager: %s", bluealsa_dbus_manager_path);
 
-	bluealsa_ManagerIfaceSkeleton *ifs_manager;
-	ifs_manager = bluealsa_manager_iface_skeleton_new(&vtable, NULL, NULL);
+	OrgBluealsaManager1Skeleton *ifs_manager;
+	ifs_manager = org_bluealsa_manager1_skeleton_new(&vtable, NULL, NULL);
 	g_dbus_interface_skeleton_export(G_DBUS_INTERFACE_SKELETON(ifs_manager),
 			config.dbus, bluealsa_dbus_manager_path, NULL);
 
@@ -405,13 +364,17 @@ static gboolean bluealsa_pcm_controller(GIOChannel *ch, GIOCondition condition,
 		void *userdata) {
 	(void)condition;
 
-	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
+	struct ba_transport_pcm *pcm = userdata;
+	GError *err = NULL;
 	char command[32];
 	size_t len;
 
-	switch (g_io_channel_read_chars(ch, command, sizeof(command), &len, NULL)) {
+	switch (g_io_channel_read_chars(ch, command, sizeof(command), &len, &err)) {
+	case G_IO_STATUS_AGAIN:
+		return TRUE;
 	case G_IO_STATUS_ERROR:
-		error("Couldn't read controller channel");
+		error("PCM controller read error: %s", err->message);
+		g_error_free(err);
 		return TRUE;
 	case G_IO_STATUS_NORMAL:
 		if (strncmp(command, BLUEALSA_PCM_CTRL_DRAIN, len) == 0) {
@@ -438,13 +401,11 @@ static gboolean bluealsa_pcm_controller(GIOChannel *ch, GIOCondition condition,
 		}
 		g_io_channel_flush(ch, NULL);
 		return TRUE;
-	case G_IO_STATUS_AGAIN:
-		return TRUE;
 	case G_IO_STATUS_EOF:
 		pthread_mutex_lock(&pcm->mutex);
 		ba_transport_pcm_release(pcm);
-		ba_transport_thread_signal_send(pcm->th, BA_TRANSPORT_THREAD_SIGNAL_PCM_CLOSE);
 		pthread_mutex_unlock(&pcm->mutex);
+		ba_transport_pcm_signal_send(pcm, BA_TRANSPORT_PCM_SIGNAL_CLOSE);
 		/* Check whether we've just closed the last PCM client and in
 		 * such a case schedule transport IO threads termination. */
 		ba_transport_stop_if_no_clients(pcm->t);
@@ -457,9 +418,9 @@ static gboolean bluealsa_pcm_controller(GIOChannel *ch, GIOCondition condition,
 
 static void bluealsa_pcm_open(GDBusMethodInvocation *inv, void *userdata) {
 
-	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
+	struct ba_transport_pcm *pcm = userdata;
 	const bool is_sink = pcm->mode == BA_TRANSPORT_PCM_MODE_SINK;
-	struct ba_transport_thread *th = pcm->th;
+	const enum ba_transport_profile t_profile = pcm->t->profile;
 	struct ba_transport *t = pcm->t;
 	int pcm_fds[4] = { -1, -1, -1, -1 };
 	size_t i;
@@ -469,7 +430,7 @@ static void bluealsa_pcm_open(GDBusMethodInvocation *inv, void *userdata) {
 	pthread_mutex_lock(&pcm->client_mtx);
 
 	/* preliminary check whether HFP codes is selected */
-	if (t->profile & BA_TRANSPORT_PROFILE_MASK_SCO &&
+	if (t_profile & BA_TRANSPORT_PROFILE_MASK_SCO &&
 			ba_transport_get_codec(t) == HFP_CODEC_UNDEFINED) {
 		g_dbus_method_invocation_return_error(inv, G_DBUS_ERROR,
 				G_DBUS_ERROR_FAILED, "HFP audio codec not selected");
@@ -482,7 +443,7 @@ static void bluealsa_pcm_open(GDBusMethodInvocation *inv, void *userdata) {
 
 	if (pcm_fd != -1) {
 		g_dbus_method_invocation_return_error(inv, G_DBUS_ERROR,
-				G_DBUS_ERROR_FAILED, "%s", strerror(EBUSY));
+				G_DBUS_ERROR_LIMITS_EXCEEDED, "%s", strerror(EBUSY));
 		goto fail;
 	}
 
@@ -506,8 +467,8 @@ static void bluealsa_pcm_open(GDBusMethodInvocation *inv, void *userdata) {
 	 * headset will not run voltage converter (power-on its circuit board) until
 	 * the transport is acquired in order to extend battery life. For profiles
 	 * like A2DP Sink and HFP headset, we will wait for incoming connection. */
-	if (t->profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE ||
-			t->profile & BA_TRANSPORT_PROFILE_MASK_AG) {
+	if (t_profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE ||
+			t_profile & BA_TRANSPORT_PROFILE_MASK_AG) {
 
 		if (ba_transport_acquire(t) == -1) {
 			g_dbus_method_invocation_return_error(inv, G_DBUS_ERROR,
@@ -516,7 +477,7 @@ static void bluealsa_pcm_open(GDBusMethodInvocation *inv, void *userdata) {
 		}
 
 		/* Wait until transport thread is ready to process audio. */
-		if (ba_transport_thread_state_wait_running(th) == -1) {
+		if (ba_transport_pcm_state_wait_running(pcm) == -1) {
 			g_dbus_method_invocation_return_error(inv, G_DBUS_ERROR,
 					G_DBUS_ERROR_IO_ERROR, "Acquire transport: %s", strerror(errno));
 			goto fail;
@@ -525,23 +486,26 @@ static void bluealsa_pcm_open(GDBusMethodInvocation *inv, void *userdata) {
 	}
 
 	pthread_mutex_lock(&pcm->mutex);
+
 	/* get correct PIPE endpoint - PIPE is unidirectional */
 	pcm->fd = pcm_fds[is_sink ? 0 : 1];
 	/* set newly opened PCM as active */
-	pcm->active = true;
-	pthread_mutex_unlock(&pcm->mutex);
+	pcm->paused = false;
 
 	GIOChannel *ch = g_io_channel_unix_new(pcm_fds[2]);
 	g_io_channel_set_close_on_unref(ch, TRUE);
 	g_io_channel_set_encoding(ch, NULL, NULL);
+	g_io_channel_set_buffered(ch, FALSE);
 
-	g_io_add_watch_full(ch, G_PRIORITY_DEFAULT, G_IO_IN,
-			bluealsa_pcm_controller, ba_transport_pcm_ref(pcm),
+	pcm->controller = g_io_create_watch_full(ch, G_PRIORITY_DEFAULT,
+			G_IO_IN, bluealsa_pcm_controller, ba_transport_pcm_ref(pcm),
 			(GDestroyNotify)ba_transport_pcm_unref);
 	g_io_channel_unref(ch);
 
-	/* notify our audio thread that the FIFO is ready */
-	ba_transport_thread_signal_send(th, BA_TRANSPORT_THREAD_SIGNAL_PCM_OPEN);
+	pthread_mutex_unlock(&pcm->mutex);
+
+	/* notify our PCM IO thread that the PCM was opened */
+	ba_transport_pcm_signal_send(pcm, BA_TRANSPORT_PCM_SIGNAL_OPEN);
 
 	int fds[2] = { pcm_fds[is_sink ? 1 : 0], pcm_fds[3] };
 	GUnixFDList *fd_list = g_unix_fd_list_new_from_array(fds, 2);
@@ -562,7 +526,7 @@ fail:
 
 static void bluealsa_pcm_get_codecs(GDBusMethodInvocation *inv, void *userdata) {
 
-	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
+	struct ba_transport_pcm *pcm = userdata;
 	const struct ba_transport *t = pcm->t;
 	const GArray *seps = t->d->seps;
 
@@ -571,7 +535,7 @@ static void bluealsa_pcm_get_codecs(GDBusMethodInvocation *inv, void *userdata)
 
 	if (t->profile & BA_TRANSPORT_PROFILE_MASK_A2DP) {
 
-		GArray *codec_ids = g_array_sized_new(FALSE, FALSE, sizeof(uint16_t), 16);
+		GArray *codec_ids = g_array_sized_new(FALSE, FALSE, sizeof(uint32_t), 16);
 		size_t i;
 
 		for (i = 0; seps != NULL && i < seps->len; i++) {
@@ -583,7 +547,7 @@ static void bluealsa_pcm_get_codecs(GDBusMethodInvocation *inv, void *userdata)
 				size_t j;
 
 				for (j = 0; j < codec_ids->len; j++)
-					if (sep->codec_id == g_array_index(codec_ids, uint16_t, j)) {
+					if (sep->codec_id == g_array_index(codec_ids, uint32_t, j)) {
 						duplicate = true;
 						break;
 					}
@@ -619,19 +583,33 @@ static void bluealsa_pcm_get_codecs(GDBusMethodInvocation *inv, void *userdata)
 		 * application like oFono, we will mark given codec as available, if it
 		 * is enabled by our global configuration. */
 
-		if ((t_sco_rfcomm != NULL &&
-					t_sco_rfcomm->ag_codecs.cvsd && t_sco_rfcomm->hf_codecs.cvsd) ||
-				(t_sco_rfcomm == NULL && config.hfp.codecs.cvsd))
-			g_variant_builder_add(&codecs, "{sa{sv}}",
-					hfp_codec_id_to_string(HFP_CODEC_CVSD), NULL);
-
+		const struct {
+			uint8_t codec_id;
+			bool is_enabled_in_config;
+			bool is_available_in_rfcomm_ag;
+			bool is_available_in_rfcomm_hf;
+		} sco_codecs[] = {
+			{ HFP_CODEC_CVSD, config.hfp.codecs.cvsd,
+				t_sco_rfcomm == NULL || t_sco_rfcomm->ag_codecs.cvsd,
+				t_sco_rfcomm == NULL || t_sco_rfcomm->hf_codecs.cvsd },
 #if ENABLE_MSBC
-		if ((t_sco_rfcomm != NULL &&
-					t_sco_rfcomm->ag_codecs.msbc && t_sco_rfcomm->hf_codecs.msbc) ||
-				(t_sco_rfcomm == NULL && config.hfp.codecs.msbc))
-			g_variant_builder_add(&codecs, "{sa{sv}}",
-					hfp_codec_id_to_string(HFP_CODEC_MSBC), NULL);
+			{ HFP_CODEC_MSBC, config.hfp.codecs.msbc,
+				t_sco_rfcomm == NULL || t_sco_rfcomm->ag_codecs.msbc,
+				t_sco_rfcomm == NULL || t_sco_rfcomm->hf_codecs.msbc },
 #endif
+#if ENABLE_LC3_SWB
+			{ HFP_CODEC_LC3_SWB, config.hfp.codecs.lc3_swb,
+				t_sco_rfcomm == NULL || t_sco_rfcomm->ag_codecs.lc3_swb,
+				t_sco_rfcomm == NULL || t_sco_rfcomm->hf_codecs.lc3_swb },
+#endif
+		};
+
+		for (size_t i = 0; i < ARRAYSIZE(sco_codecs); i++)
+			if (sco_codecs[i].is_enabled_in_config &&
+					sco_codecs[i].is_available_in_rfcomm_ag &&
+					sco_codecs[i].is_available_in_rfcomm_hf)
+				g_variant_builder_add(&codecs, "{sa{sv}}",
+						hfp_codec_id_to_string(sco_codecs[i].codec_id), NULL);
 
 	}
 
@@ -643,7 +621,7 @@ static void bluealsa_pcm_get_codecs(GDBusMethodInvocation *inv, void *userdata)
 static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv, void *userdata) {
 
 	GVariant *params = g_dbus_method_invocation_get_parameters(inv);
-	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
+	struct ba_transport_pcm *pcm = userdata;
 	struct ba_transport *t = pcm->t;
 	GVariantIter *properties;
 	GVariant *value = NULL;
@@ -659,6 +637,7 @@ static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv, void *userdata
 
 	a2dp_t a2dp_configuration = {};
 	size_t a2dp_configuration_size = 0;
+	bool conformance_check = true;
 
 	g_variant_get(params, "(&sa{sv})", &codec_name, &properties);
 	while (g_variant_iter_next(properties, "{&sv}", &property, &value)) {
@@ -678,6 +657,10 @@ static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv, void *userdata
 			memcpy(&a2dp_configuration, data, a2dp_configuration_size);
 
 		}
+		else if (strcmp(property, "NonConformant") == 0 &&
+				g_variant_validate_value(value, G_VARIANT_TYPE_BOOLEAN, property)) {
+			conformance_check = !g_variant_get_boolean(value);
+		}
 
 		g_variant_unref(value);
 		value = NULL;
@@ -691,7 +674,7 @@ static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv, void *userdata
 			goto fail;
 		}
 
-		uint16_t codec_id = a2dp_codecs_codec_id_from_string(codec_name);
+		uint32_t codec_id = a2dp_codecs_codec_id_from_string(codec_name);
 		enum a2dp_dir dir = !t->a2dp.codec->dir;
 		const GArray *seps = t->d->seps;
 		struct a2dp_sep *sep = NULL;
@@ -716,19 +699,42 @@ static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv, void *userdata
 			goto fail;
 		}
 
-		/* setup default codec configuration */
-		memcpy(&sep->configuration, &sep->capabilities, sep->capabilities_size);
-		if (a2dp_select_configuration(codec, &sep->configuration, sep->capabilities_size) == -1)
-			goto fail;
+		if (a2dp_configuration_size == 0) {
+			/* setup default codec configuration */
 
-		/* use codec configuration blob provided by user */
-		if (a2dp_configuration_size != 0) {
-			if (a2dp_check_configuration(codec, &a2dp_configuration,
-						a2dp_configuration_size) != A2DP_CHECK_OK) {
-				errmsg = "Invalid configuration blob";
+			const size_t size = sep->capabilities_size;
+			memcpy(&sep->configuration, &sep->capabilities, size);
+			if (a2dp_select_configuration(codec, &sep->configuration, size) == -1)
 				goto fail;
+
+		}
+		else {
+			/* use codec configuration blob provided by user */
+
+			if (conformance_check) {
+
+				a2dp_filter_capabilities(codec, &sep->capabilities,
+						&a2dp_configuration, a2dp_configuration_size);
+
+				enum a2dp_check_err rv;
+				if ((rv = a2dp_check_configuration(codec, &a2dp_configuration,
+							a2dp_configuration_size)) != A2DP_CHECK_OK) {
+					errmsg = a2dp_check_strerror(rv);
+					goto fail;
+				}
+
 			}
-			memcpy(&sep->configuration, &a2dp_configuration, sep->capabilities_size);
+			else {
+
+				if (a2dp_configuration_size != sep->capabilities_size) {
+					errmsg = a2dp_check_strerror(A2DP_CHECK_ERR_SIZE);
+					goto fail;
+				}
+
+			}
+
+			memcpy(&sep->configuration, &a2dp_configuration, a2dp_configuration_size);
+
 		}
 
 		if (ba_transport_select_codec_a2dp(t, sep) == -1)
@@ -737,7 +743,7 @@ static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv, void *userdata
 	}
 	else {
 
-		uint16_t codec_id;
+		uint8_t codec_id;
 		if ((codec_id = hfp_codec_id_from_string(codec_name)) == HFP_CODEC_UNDEFINED) {
 			errmsg = "HFP codec not available";
 			goto fail;
@@ -765,14 +771,82 @@ final:
 		g_variant_unref(value);
 }
 
+static void bluealsa_pcm_set_delay_adjustment(GDBusMethodInvocation *inv, void *userdata) {
+
+	GVariant *params = g_dbus_method_invocation_get_parameters(inv);
+	struct ba_transport_pcm *pcm = userdata;
+	const struct ba_transport *t = pcm->t;
+
+	const char *codec;
+	int16_t adjustment;
+
+	g_variant_get(params, "(&sn)", &codec, &adjustment);
+
+	uint32_t codec_id = 0;
+	bool is_valid = false;
+	if (t->profile & BA_TRANSPORT_PROFILE_MASK_A2DP) {
+		codec_id = a2dp_codecs_codec_id_from_string(codec);
+		is_valid = codec_id != 0xFFFFFFFF;
+	}
+	if (t->profile & BA_TRANSPORT_PROFILE_MASK_SCO) {
+		codec_id = hfp_codec_id_from_string(codec);
+		is_valid = codec_id != HFP_CODEC_UNDEFINED;
+	}
+
+	if (!is_valid) {
+		error("Invalid codec name: %s", codec);
+		g_dbus_method_invocation_return_error(inv, G_DBUS_ERROR,
+				G_DBUS_ERROR_INVALID_ARGS, "Invalid codec name: %s", codec);
+		return;
+	}
+
+	ba_transport_pcm_delay_adjustment_set(pcm, codec_id, adjustment);
+	bluealsa_dbus_pcm_update(pcm, BA_DBUS_PCM_UPDATE_DELAY_ADJUSTMENT);
+	g_dbus_method_invocation_return_value(inv, NULL);
+
+}
+
+static void bluealsa_pcm_get_delay_adjustments(GDBusMethodInvocation *inv, void *userdata) {
+
+	struct ba_transport_pcm *pcm = userdata;
+	const struct ba_transport *t = pcm->t;
+
+	GVariantBuilder adjustments;
+	g_variant_builder_init(&adjustments, G_VARIANT_TYPE("a{sn}"));
+
+	pthread_mutex_lock(&pcm->delay_adjustments_mtx);
+
+	GHashTableIter iter;
+	g_hash_table_iter_init(&iter, pcm->delay_adjustments);
+
+	void *key, *value;
+	while (g_hash_table_iter_next(&iter, &key, &value)) {
+		const char *codec = NULL;
+		if (t->profile & BA_TRANSPORT_PROFILE_MASK_A2DP)
+			codec = a2dp_codecs_codec_id_to_string(GPOINTER_TO_INT(key));
+		if (t->profile & BA_TRANSPORT_PROFILE_MASK_SCO)
+			codec = hfp_codec_id_to_string(GPOINTER_TO_INT(key));
+		if (codec != NULL) {
+			int16_t adjustment = GPOINTER_TO_INT(value);
+			g_variant_builder_add(&adjustments, "{sn}", codec, adjustment);
+		}
+	}
+
+	pthread_mutex_unlock(&pcm->delay_adjustments_mtx);
+
+	g_dbus_method_invocation_return_value(inv, g_variant_new("(a{sn})", &adjustments));
+	g_variant_builder_clear(&adjustments);
+
+}
+
 static void bluealsa_rfcomm_open(GDBusMethodInvocation *inv, void *userdata) {
 
-	struct ba_rfcomm *r = (struct ba_rfcomm *)userdata;
+	struct ba_rfcomm *r = userdata;
 	int fds[2] = { -1, -1 };
 
 	if (r->handler_fd != -1) {
 		g_dbus_method_invocation_return_error(inv, G_DBUS_ERROR,
-				G_DBUS_ERROR_FAILED, "%s", strerror(EBUSY));
+				G_DBUS_ERROR_LIMITS_EXCEEDED, "%s", strerror(EBUSY));
 		return;
 	}
 
@@ -791,21 +865,10 @@ static void bluealsa_rfcomm_open(GDBusMethodInvocation *inv, void *userdata) {
 	g_object_unref(fd_list);
 }
 
-static GVariant *bluealsa_pcm_get_properties(void *userdata) {
-
-	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
-
-	GVariantBuilder props;
-
-	ba_variant_populate_pcm(&props, pcm);
-
-	return g_variant_builder_end(&props);
-}
-
 static GVariant *bluealsa_pcm_get_property(const char *property,
 		GError **error, void *userdata) {
 
-	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
+	struct ba_transport_pcm *pcm = userdata;
 	struct ba_device *d = pcm->t->d;
 	GVariant *value;
 
@@ -837,6 +900,8 @@ static GVariant *bluealsa_pcm_get_property(const char *property,
 	}
 	if (strcmp(property, "Delay") == 0)
 		return ba_variant_new_pcm_delay(pcm);
+	if (strcmp(property, "DelayAdjustment") == 0)
+		return ba_variant_new_pcm_delay_adjustment(pcm);
 	if (strcmp(property, "SoftVolume") == 0)
 		return ba_variant_new_pcm_soft_volume(pcm);
 	if (strcmp(property, "Volume") == 0)
@@ -856,7 +921,7 @@ static bool bluealsa_pcm_set_property(const char *property, GVariant *value,
 		GError **error, void *userdata) {
 	(void)error;
 
-	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
+	struct ba_transport_pcm *pcm = userdata;
 
 	if (strcmp(property, "SoftVolume") == 0) {
 		pcm->soft_volume = g_variant_get_boolean(value);
@@ -906,26 +971,32 @@ int bluealsa_dbus_pcm_register(struct ba_transport_pcm *pcm) {
 			.handler = bluealsa_pcm_get_codecs },
 		{ .method = "SelectCodec",
 			.handler = bluealsa_pcm_select_codec },
+		{ .method = "SetDelayAdjustment",
+			.handler = bluealsa_pcm_set_delay_adjustment },
+		{ .method = "GetDelayAdjustments",
+			.handler = bluealsa_pcm_get_delay_adjustments },
 		{ 0 },
 	};
 
 	static const GDBusInterfaceSkeletonVTable vtable = {
 		.dispatchers = dispatchers,
-		.get_properties = bluealsa_pcm_get_properties,
 		.get_property = bluealsa_pcm_get_property,
 		.set_property = bluealsa_pcm_set_property,
 	};
 
 	GDBusObjectSkeleton *skeleton = NULL;
-	bluealsa_PCMIfaceSkeleton *ifs_pcm = NULL;
+	OrgBluealsaPcm1Skeleton *ifs_pcm = NULL;
 
 	if ((skeleton = g_dbus_object_skeleton_new(pcm->ba_dbus_path)) == NULL)
 		goto fail;
 
-	if ((ifs_pcm = bluealsa_pcm_iface_skeleton_new(&vtable,
+	if ((ifs_pcm = org_bluealsa_pcm1_skeleton_new(&vtable,
 					pcm, (GDestroyNotify)ba_transport_pcm_unref)) == NULL)
 		goto fail;
 
+	g_dbus_interface_skeleton_set_flags(G_DBUS_INTERFACE_SKELETON(ifs_pcm),
+			G_DBUS_INTERFACE_SKELETON_FLAGS_HANDLE_METHOD_INVOCATIONS_IN_THREAD);
+
 	ba_transport_pcm_ref(pcm);
 
 	g_dbus_object_skeleton_add_interface(skeleton, G_DBUS_INTERFACE_SKELETON(ifs_pcm));
@@ -959,16 +1030,17 @@ void bluealsa_dbus_pcm_update(struct ba_transport_pcm *pcm, unsigned int mask) {
 		g_variant_builder_add(&props, "{sv}", "Codec", ba_variant_new_pcm_codec(pcm));
 	if (mask & BA_DBUS_PCM_UPDATE_CODEC_CONFIG)
 		g_variant_builder_add(&props, "{sv}", "CodecConfiguration", ba_variant_new_pcm_codec_config(pcm));
-	if (mask & BA_DBUS_PCM_UPDATE_DELAY)
+	if (mask & (BA_DBUS_PCM_UPDATE_DELAY | BA_DBUS_PCM_UPDATE_DELAY_ADJUSTMENT))
 		g_variant_builder_add(&props, "{sv}", "Delay", ba_variant_new_pcm_delay(pcm));
+	if (mask & BA_DBUS_PCM_UPDATE_DELAY_ADJUSTMENT)
+		g_variant_builder_add(&props, "{sv}", "DelayAdjustment", ba_variant_new_pcm_delay_adjustment(pcm));
 	if (mask & BA_DBUS_PCM_UPDATE_SOFT_VOLUME)
 		g_variant_builder_add(&props, "{sv}", "SoftVolume", ba_variant_new_pcm_soft_volume(pcm));
 	if (mask & BA_DBUS_PCM_UPDATE_VOLUME)
 		g_variant_builder_add(&props, "{sv}", "Volume", ba_variant_new_pcm_volume(pcm));
 
-	g_dbus_connection_emit_properties_changed(config.dbus,
-			pcm->ba_dbus_path, BLUEALSA_IFACE_PCM, &props, NULL);
-	g_variant_builder_clear(&props);
+	g_dbus_connection_emit_properties_changed(config.dbus, pcm->ba_dbus_path,
+			BLUEALSA_IFACE_PCM, g_variant_builder_end(&props), NULL, NULL);
 
 }
 
@@ -982,27 +1054,11 @@ void bluealsa_dbus_pcm_unregister(struct ba_transport_pcm *pcm) {
 
 }
 
-static GVariant *bluealsa_rfcomm_get_properties(void *userdata) {
-
-	struct ba_rfcomm *r = (struct ba_rfcomm *)userdata;
-	struct ba_transport *t = r->sco;
-	struct ba_device *d = t->d;
-
-	GVariantBuilder props;
-	g_variant_builder_init(&props, G_VARIANT_TYPE("a{sv}"));
-
-	g_variant_builder_add(&props, "{sv}", "Transport", ba_variant_new_transport_type(t));
-	g_variant_builder_add(&props, "{sv}", "Features", ba_variant_new_rfcomm_features(r));
-	g_variant_builder_add(&props, "{sv}", "Battery", ba_variant_new_device_battery(d));
-
-	return g_variant_builder_end(&props);
-}
-
 static GVariant *bluealsa_rfcomm_get_property(const char *property,
 		GError **error, void *userdata) {
 	(void)error;
 
-	struct ba_rfcomm *r = (struct ba_rfcomm *)userdata;
+	struct ba_rfcomm *r = userdata;
 	struct ba_transport *t = r->sco;
 	struct ba_device *d = t->d;
 
@@ -1029,17 +1085,16 @@ int bluealsa_dbus_rfcomm_register(struct ba_rfcomm *r) {
 
 	static const GDBusInterfaceSkeletonVTable vtable = {
 		.dispatchers = dispatchers,
-		.get_properties = bluealsa_rfcomm_get_properties,
 		.get_property = bluealsa_rfcomm_get_property,
 	};
 
 	GDBusObjectSkeleton *skeleton = NULL;
-	bluealsa_RFCOMMIfaceSkeleton *ifs_rfcomm = NULL;
+	OrgBluealsaRfcomm1Skeleton *ifs_rfcomm = NULL;
 
 	if ((skeleton = g_dbus_object_skeleton_new(r->ba_dbus_path)) == NULL)
 		goto fail;
 
-	if ((ifs_rfcomm = bluealsa_rfcomm_iface_skeleton_new(&vtable,
+	if ((ifs_rfcomm = org_bluealsa_rfcomm1_skeleton_new(&vtable,
 					r, NULL)) == NULL)
 		goto fail;
 
@@ -1067,9 +1122,8 @@ void bluealsa_dbus_rfcomm_update(struct ba_rfcomm *r, unsigned int mask) {
 	if (mask & BA_DBUS_RFCOMM_UPDATE_BATTERY)
 		g_variant_builder_add(&props, "{sv}", "Battery", ba_variant_new_device_battery(r->sco->d));
 
-	g_dbus_connection_emit_properties_changed(config.dbus,
-			r->ba_dbus_path, BLUEALSA_IFACE_RFCOMM, &props, NULL);
-	g_variant_builder_clear(&props);
+	g_dbus_connection_emit_properties_changed(config.dbus, r->ba_dbus_path,
+			BLUEALSA_IFACE_RFCOMM, g_variant_builder_end(&props), NULL, NULL);
 
 }
 
