@@ -1,6 +1,6 @@
 /*
  * test-io.c
- * Copyright (c) 2016-2023 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -23,6 +23,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <strings.h>
+#include <sys/ioctl.h>
 #include <sys/socket.h>
 #include <unistd.h>
 
@@ -61,14 +62,16 @@
 #endif
 #include "a2dp-sbc.h"
 #include "ba-adapter.h"
+#include "ba-config.h"
 #include "ba-device.h"
 #include "ba-rfcomm.h"
 #include "ba-transport.h"
-#include "bluealsa-config.h"
+#include "ba-transport-pcm.h"
 #include "bluealsa-dbus.h"
 #include "bluez.h"
 #include "hfp.h"
 #include "io.h"
+#include "midi.h"
 #if ENABLE_OFONO
 # include "ofono.h"
 #endif
@@ -91,24 +94,24 @@
 		(CHECK_MINOR_VERSION << 8 & 0x00ff00) | \
 		(CHECK_MICRO_VERSION << 0 & 0x0000ff))
 
-void *a2dp_aac_dec_thread(struct ba_transport_thread *th);
-void *a2dp_aac_enc_thread(struct ba_transport_thread *th);
-void *a2dp_aptx_dec_thread(struct ba_transport_thread *th);
-void *a2dp_aptx_enc_thread(struct ba_transport_thread *th);
-void *a2dp_aptx_hd_dec_thread(struct ba_transport_thread *th);
-void *a2dp_aptx_hd_enc_thread(struct ba_transport_thread *th);
-void *a2dp_faststream_dec_thread(struct ba_transport_thread *th);
-void *a2dp_faststream_enc_thread(struct ba_transport_thread *th);
-void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th);
-void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th);
-void *a2dp_ldac_dec_thread(struct ba_transport_thread *th);
-void *a2dp_ldac_enc_thread(struct ba_transport_thread *th);
-void *a2dp_mp3_enc_thread(struct ba_transport_thread *th);
-void *a2dp_mpeg_dec_thread(struct ba_transport_thread *th);
-void *a2dp_sbc_dec_thread(struct ba_transport_thread *th);
-void *a2dp_sbc_enc_thread(struct ba_transport_thread *th);
-void *sco_dec_thread(struct ba_transport_thread *th);
-void *sco_enc_thread(struct ba_transport_thread *th);
+void *a2dp_aac_dec_thread(struct ba_transport_pcm *t_pcm);
+void *a2dp_aac_enc_thread(struct ba_transport_pcm *t_pcm);
+void *a2dp_aptx_dec_thread(struct ba_transport_pcm *t_pcm);
+void *a2dp_aptx_enc_thread(struct ba_transport_pcm *t_pcm);
+void *a2dp_aptx_hd_dec_thread(struct ba_transport_pcm *t_pcm);
+void *a2dp_aptx_hd_enc_thread(struct ba_transport_pcm *t_pcm);
+void *a2dp_faststream_dec_thread(struct ba_transport_pcm *t_pcm);
+void *a2dp_faststream_enc_thread(struct ba_transport_pcm *t_pcm);
+void *a2dp_lc3plus_dec_thread(struct ba_transport_pcm *t_pcm);
+void *a2dp_lc3plus_enc_thread(struct ba_transport_pcm *t_pcm);
+void *a2dp_ldac_dec_thread(struct ba_transport_pcm *t_pcm);
+void *a2dp_ldac_enc_thread(struct ba_transport_pcm *t_pcm);
+void *a2dp_mp3_enc_thread(struct ba_transport_pcm *t_pcm);
+void *a2dp_mpeg_dec_thread(struct ba_transport_pcm *t_pcm);
+void *a2dp_sbc_dec_thread(struct ba_transport_pcm *t_pcm);
+void *a2dp_sbc_enc_thread(struct ba_transport_pcm *t_pcm);
+void *sco_dec_thread(struct ba_transport_pcm *t_pcm);
+void *sco_enc_thread(struct ba_transport_pcm *t_pcm);
 
 int bluealsa_dbus_pcm_register(struct ba_transport_pcm *pcm) {
 	debug("%s: %p", __func__, (void *)pcm); (void)pcm; return 0; }
@@ -128,6 +131,10 @@ bool bluez_a2dp_set_configuration(const char *current_dbus_sep_path,
 	(void)current_dbus_sep_path; (void)sep; (void)error; return false; }
 int ofono_call_volume_update(struct ba_transport *t) {
 	debug("%s: %p", __func__, t); (void)t; return 0; }
+int midi_transport_alsa_seq_create(struct ba_transport *t) { (void)t; return 0; }
+int midi_transport_alsa_seq_delete(struct ba_transport *t) { (void)t; return 0; }
+int midi_transport_start(struct ba_transport *t) { (void)t; return 0; }
+int midi_transport_stop(struct ba_transport *t) { (void)t; return 0; }
 int storage_device_load(const struct ba_device *d) { (void)d; return 0; }
 int storage_device_save(const struct ba_device *d) { (void)d; return 0; }
 int storage_pcm_data_sync(struct ba_transport_pcm *pcm) { (void)pcm; return 0; }
@@ -148,34 +155,34 @@ static a2dp_mpeg_t config_mp3_44100_stereo = {
 	.layer = MPEG_LAYER_MP3,
 	.channel_mode = MPEG_CHANNEL_MODE_STEREO,
 	.frequency = MPEG_SAMPLING_FREQ_44100,
-	MPEG_INIT_BITRATE(0xFFFF)
+	A2DP_MPEG_INIT_BITRATE(0xFFFF)
 };
 
 __attribute__ ((unused))
 static a2dp_aac_t config_aac_44100_stereo = {
-	.object_type = AAC_OBJECT_TYPE_MPEG2_AAC_LC,
-	AAC_INIT_FREQUENCY(AAC_SAMPLING_FREQ_44100)
+	.object_type = AAC_OBJECT_TYPE_MPEG2_LC,
+	A2DP_AAC_INIT_FREQUENCY(AAC_SAMPLING_FREQ_44100)
 	.channels = AAC_CHANNELS_2,
-	AAC_INIT_BITRATE(0xFFFF)
+	A2DP_AAC_INIT_BITRATE(0xFFFF)
 };
 
 __attribute__ ((unused))
 static const a2dp_aptx_t config_aptx_44100_stereo = {
-	.info = A2DP_SET_VENDOR_ID_CODEC_ID(APTX_VENDOR_ID, APTX_CODEC_ID),
+	.info = A2DP_VENDOR_INFO_INIT(APTX_VENDOR_ID, APTX_CODEC_ID),
 	.frequency = APTX_SAMPLING_FREQ_44100,
 	.channel_mode = APTX_CHANNEL_MODE_STEREO,
 };
 
 __attribute__ ((unused))
 static const a2dp_aptx_hd_t config_aptx_hd_44100_stereo = {
-	.aptx.info = A2DP_SET_VENDOR_ID_CODEC_ID(APTX_HD_VENDOR_ID, APTX_HD_CODEC_ID),
+	.aptx.info = A2DP_VENDOR_INFO_INIT(APTX_HD_VENDOR_ID, APTX_HD_CODEC_ID),
 	.aptx.frequency = APTX_SAMPLING_FREQ_44100,
 	.aptx.channel_mode = APTX_CHANNEL_MODE_STEREO,
 };
 
 __attribute__ ((unused))
 static const a2dp_faststream_t config_faststream_44100_16000 = {
-	.info = A2DP_SET_VENDOR_ID_CODEC_ID(FASTSTREAM_VENDOR_ID, FASTSTREAM_CODEC_ID),
+	.info = A2DP_VENDOR_INFO_INIT(FASTSTREAM_VENDOR_ID, FASTSTREAM_CODEC_ID),
 	.direction = FASTSTREAM_DIRECTION_MUSIC | FASTSTREAM_DIRECTION_VOICE,
 	.frequency_music = FASTSTREAM_SAMPLING_FREQ_MUSIC_44100,
 	.frequency_voice = FASTSTREAM_SAMPLING_FREQ_VOICE_16000,
@@ -183,15 +190,15 @@ static const a2dp_faststream_t config_faststream_44100_16000 = {
 
 __attribute__ ((unused))
 static const a2dp_lc3plus_t config_lc3plus_48000_stereo = {
-	.info = A2DP_SET_VENDOR_ID_CODEC_ID(LC3PLUS_VENDOR_ID, LC3PLUS_CODEC_ID),
+	.info = A2DP_VENDOR_INFO_INIT(LC3PLUS_VENDOR_ID, LC3PLUS_CODEC_ID),
 	.frame_duration = LC3PLUS_FRAME_DURATION_050,
 	.channels = LC3PLUS_CHANNELS_2,
-	LC3PLUS_INIT_FREQUENCY(LC3PLUS_SAMPLING_FREQ_48000)
+	A2DP_LC3PLUS_INIT_FREQUENCY(LC3PLUS_SAMPLING_FREQ_48000)
 };
 
 __attribute__ ((unused))
 static const a2dp_ldac_t config_ldac_44100_stereo = {
-	.info = A2DP_SET_VENDOR_ID_CODEC_ID(LDAC_VENDOR_ID, LDAC_CODEC_ID),
+	.info = A2DP_VENDOR_INFO_INIT(LDAC_VENDOR_ID, LDAC_CODEC_ID),
 	.frequency = LDAC_SAMPLING_FREQ_44100,
 	.channel_mode = LDAC_CHANNEL_MODE_STEREO,
 };
@@ -260,7 +267,7 @@ static void *pcm_write_frames_sndfile_async(void *userdata) {
 	};
 
 	if (aging_duration == 0) {
-		/* If we are not performign aging test, close the PCM right
+		/* If we are not performing aging test, close the PCM right
 		 * away. The reading loop will not wait for timeout. */
 		pthread_mutex_lock(&pcm->mutex);
 		ba_transport_pcm_release(pcm);
@@ -328,7 +335,7 @@ static void pcm_write_frames(struct ba_transport_pcm *pcm, size_t frames) {
 	}
 
 	if (aging_duration == 0) {
-		/* If we are not performign aging test, close the PCM right
+		/* If we are not performing aging test, close the PCM right
 		 * away. The reading loop will not wait for timeout. */
 		pthread_mutex_lock(&pcm->mutex);
 		ba_transport_pcm_release(pcm);
@@ -379,6 +386,11 @@ static void bt_data_write(struct ba_transport *t) {
 			ck_assert_int_ne(poll(fds, ARRAYSIZE(fds), -1), -1);
 			ck_assert_int_eq(write(fds[0].fd, buffer, len), len);
 			first_packet = false;
+			/* Allow the test_io_thread_dump_pcm() to read out the decoded PCM
+			 * frames. Otherwise, the PCM FIFO might get full and the decoder
+			 * will detect an overrun. In such case, the decoder will drop the
+			 * PCM frames and we will not decode the whole BT dump. */
+			usleep(500);
 		}
 
 	}
@@ -393,6 +405,7 @@ static void bt_data_write(struct ba_transport *t) {
 			ck_assert_int_ne(poll(fds, ARRAYSIZE(fds), -1), -1);
 			ck_assert_int_eq(write(fds[0].fd, bt_data_head->data, len), len);
 			first_packet = false;
+			usleep(500);
 		}
 
 	}
@@ -414,12 +427,12 @@ static void test_start_terminate_timer(unsigned int delay) {
 	pthread_detach(thread);
 }
 
-static void *test_io_thread_dump_bt(struct ba_transport_thread *th) {
+static void *test_io_thread_dump_bt(struct ba_transport_pcm *t_pcm) {
 
-	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
+	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_pcm_thread_cleanup), t_pcm);
 
-	struct ba_transport *t = th->t;
-	struct pollfd pfds[] = {{ th->bt_fd, POLLIN, 0 }};
+	struct ba_transport *t = t_pcm->t;
+	struct pollfd pfds[] = {{ t_pcm->fd_bt, POLLIN, 0 }};
 	struct bt_dump *btd = NULL;
 	uint8_t buffer[1024];
 	ssize_t len;
@@ -430,11 +443,11 @@ static void *test_io_thread_dump_bt(struct ba_transport_thread *th) {
 		ck_assert_ptr_ne(btd = bt_dump_create(fname, t), NULL);
 	}
 
-	debug_transport_thread_loop(th, "START");
-	ba_transport_thread_state_set_running(th);
+	debug_transport_pcm_thread_loop(t_pcm, "START");
+	ba_transport_pcm_state_set_running(t_pcm);
 	while (poll(pfds, ARRAYSIZE(pfds), 500) > 0) {
 
-		if ((len = io_bt_read(th, buffer, sizeof(buffer))) <= 0) {
+		if ((len = io_bt_read(t_pcm, buffer, sizeof(buffer))) <= 0) {
 			if (len == -1)
 				error("BT read error: %s", strerror(errno));
 			break;
@@ -459,11 +472,10 @@ static void *test_io_thread_dump_bt(struct ba_transport_thread *th) {
 	return NULL;
 }
 
-static void *test_io_thread_dump_pcm(struct ba_transport_thread *th) {
+static void *test_io_thread_dump_pcm(struct ba_transport_pcm *t_pcm) {
 
-	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
+	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_pcm_thread_cleanup), t_pcm);
 
-	struct ba_transport_pcm *t_pcm = th->pcm;
 	size_t decoded_samples_total = 0;
 
 #if HAVE_SNDFILE
@@ -494,15 +506,15 @@ static void *test_io_thread_dump_pcm(struct ba_transport_thread *th) {
 	if (dump_data) {
 #if HAVE_SNDFILE
 		char fname[64];
-		sprintf(fname, "decoded-%s.wav", transport_to_fname(th->t));
+		sprintf(fname, "decoded-%s.wav", transport_to_fname(t_pcm->t));
 		ck_assert_ptr_ne(sf = sf_open(fname, SFM_WRITE, &sf_info), NULL);
 #else
 		error("Dumping audio files requires sndfile library!");
 #endif
 	}
 
-	debug_transport_thread_loop(th, "START");
-	for (ba_transport_thread_state_set_running(th);;) {
+	debug_transport_pcm_thread_loop(t_pcm, "START");
+	for (ba_transport_pcm_state_set_running(t_pcm);;) {
 
 		struct pollfd pfds[] = {{ -1, POLLIN, 0 }};
 		uint8_t buffer[2048];
@@ -564,21 +576,21 @@ static void *test_io_thread_dump_pcm(struct ba_transport_thread *th) {
 
 /**
  * Drive PCM signal through source/sink loop. */
-static void test_io(struct ba_transport *t_src, struct ba_transport *t_snk,
-		void *(*enc)(struct ba_transport_thread *), void *(*dec)(struct ba_transport_thread *),
+static void test_io(
+		struct ba_transport_pcm *t_src_pcm, struct ba_transport_pcm *t_snk_pcm,
+		ba_transport_pcm_thread_func enc, ba_transport_pcm_thread_func dec,
 		size_t pcm_write_frames_count) {
 
 	const char *enc_name = enc == test_io_thread_dump_pcm ? "dump-pcm" : "encode";
 	const char *dec_name = dec == test_io_thread_dump_bt ? "dump-bt" : "decode";
+	struct ba_transport *t_src = t_src_pcm->t;
+	struct ba_transport *t_snk = t_snk_pcm->t;
 
 	if (enc == test_io_thread_dump_pcm && input_pcm_file != NULL)
 		return;
 	if (dec == test_io_thread_dump_bt && input_bt_file != NULL)
 		return;
 
-	struct ba_transport_pcm *t_src_pcm = t_src->thread_enc.pcm;
-	struct ba_transport_pcm *t_snk_pcm = t_snk->thread_dec.pcm;
-
 	int bt_fds[2];
 	ck_assert_int_eq(socketpair(AF_UNIX, SOCK_SEQPACKET | SOCK_NONBLOCK, 0, bt_fds), 0);
 	debug("Created BT socket pair: %d, %d", bt_fds[0], bt_fds[1]);
@@ -598,13 +610,13 @@ static void test_io(struct ba_transport *t_src, struct ba_transport *t_snk,
 		test_start_terminate_timer(aging_duration);
 
 	if (enc == test_io_thread_dump_pcm) {
-		ck_assert_int_eq(ba_transport_thread_create(&t_snk->thread_dec, dec, dec_name, true), 0);
-		ck_assert_int_eq(ba_transport_thread_create(&t_src->thread_enc, enc, enc_name, true), 0);
+		ck_assert_int_eq(ba_transport_pcm_start(t_snk_pcm, dec, dec_name), 0);
+		ck_assert_int_eq(ba_transport_pcm_start(t_src_pcm, enc, enc_name), 0);
 		bt_data_write(t_src);
 	}
 	else {
-		ck_assert_int_eq(ba_transport_thread_create(&t_src->thread_enc, enc, enc_name, true), 0);
-		ck_assert_int_eq(ba_transport_thread_create(&t_snk->thread_dec, dec, dec_name, true), 0);
+		ck_assert_int_eq(ba_transport_pcm_start(t_src_pcm, enc, enc_name), 0);
+		ck_assert_int_eq(ba_transport_pcm_start(t_snk_pcm, dec, dec_name), 0);
 		pcm_write_frames(t_snk_pcm, pcm_write_frames_count);
 	}
 
@@ -672,14 +684,17 @@ CK_START_TEST(test_a2dp_sbc) {
 			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/sbc", &a2dp_sbc_sink,
 			&config_sbc_44100_stereo);
 
+	struct ba_transport_pcm *t1_pcm = &t1->a2dp.pcm;
+	struct ba_transport_pcm *t2_pcm = &t2->a2dp.pcm;
+
 	if (aging_duration) {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 153 * 3;
-		test_io(t1, t2, a2dp_sbc_enc_thread, a2dp_sbc_dec_thread, 4 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_sbc_enc_thread, a2dp_sbc_dec_thread, 4 * 1024);
 	}
 	else {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 153 * 3;
-		test_io(t1, t2, a2dp_sbc_enc_thread, test_io_thread_dump_bt, 2 * 1024);
-		test_io(t1, t2, test_io_thread_dump_pcm, a2dp_sbc_dec_thread, 2 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_sbc_enc_thread, test_io_thread_dump_bt, 2 * 1024);
+		test_io(t1_pcm, t2_pcm, test_io_thread_dump_pcm, a2dp_sbc_dec_thread, 2 * 1024);
 	}
 
 	ba_transport_destroy(t1);
@@ -700,15 +715,135 @@ CK_START_TEST(test_a2dp_sbc_invalid_config) {
 	t->mtu_read = t->mtu_write = 153 * 3;
 	t->bt_fd = bt_fds[1];
 
-	struct ba_transport_thread *th = &t->thread_enc;
-	ck_assert_int_eq(ba_transport_thread_create(th, a2dp_sbc_enc_thread, "sbc", true), 0);
-	ck_assert_int_eq(ba_transport_thread_state_wait_running(th), -1);
+	ck_assert_int_eq(ba_transport_pcm_start(&t->a2dp.pcm, a2dp_sbc_enc_thread, "sbc"), 0);
+	ck_assert_int_eq(ba_transport_pcm_state_wait_running(&t->a2dp.pcm), -1);
 
 	ba_transport_destroy(t);
 	close(bt_fds[0]);
 
 } CK_END_TEST
 
+static void setup_a2dp_link(struct ba_transport *t_source, struct ba_transport *t_sink,
+		size_t mtu, int *fd_pcm_sink, int *fd_pcm_source) {
+
+	int bt_fds[2];
+	/* Link transports together using BT socket pair. */
+	ck_assert_int_eq(socketpair(AF_UNIX, SOCK_SEQPACKET | SOCK_NONBLOCK, 0, bt_fds), 0);
+	debug("Created BT socket pair: %d, %d", bt_fds[0], bt_fds[1]);
+	t_source->mtu_read = t_source->mtu_write = mtu;
+	t_sink->mtu_read = t_sink->mtu_write = mtu;
+	t_source->bt_fd = bt_fds[1];
+	t_sink->bt_fd = bt_fds[0];
+
+	int pcm_snk_fds[2];
+	/* Attach sink PCM to the source transport. */
+	ck_assert_int_eq(pipe2(pcm_snk_fds, O_NONBLOCK), 0);
+	debug("Created PCM pipe pair: %d, %d", pcm_snk_fds[0], pcm_snk_fds[1]);
+	t_source->a2dp.pcm.fd = pcm_snk_fds[0];
+	t_source->a2dp.pcm.paused = false;
+
+	int pcm_src_fds[2];
+	/* Attach source PCM to the sink transport. */
+	ck_assert_int_eq(pipe2(pcm_src_fds, O_NONBLOCK), 0);
+	debug("Created PCM pipe pair: %d, %d", pcm_src_fds[0], pcm_src_fds[1]);
+	t_sink->a2dp.pcm.fd = pcm_src_fds[1];
+	t_sink->a2dp.pcm.paused = false;
+
+	*fd_pcm_sink = pcm_snk_fds[1];
+	*fd_pcm_source = pcm_src_fds[0];
+
+}
+
+CK_START_TEST(test_a2dp_sbc_pcm_drain) {
+
+	int16_t pcm_zero[90] = { 0 };
+	unsigned int nread = 0;
+
+	struct ba_transport *t1 = test_transport_new_a2dp(device1,
+			BA_TRANSPORT_PROFILE_A2DP_SOURCE, "/path/sbc", &a2dp_sbc_source,
+			&config_sbc_44100_stereo);
+	struct ba_transport *t2 = test_transport_new_a2dp(device2,
+			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/sbc", &a2dp_sbc_sink,
+			&config_sbc_44100_stereo);
+
+	int fd_pcm_snk = -1;
+	int fd_pcm_src = -1;
+	setup_a2dp_link(t1, t2, 256, &fd_pcm_snk, &fd_pcm_src);
+
+	/* start sink PCM IO thread and make sure it is running */
+	struct ba_transport_pcm *pcm = &t1->a2dp.pcm;
+	ck_assert_int_eq(ba_transport_pcm_start(pcm, a2dp_sbc_enc_thread, "sbc"), 0);
+	ck_assert_int_eq(ba_transport_pcm_state_wait_running(pcm), 0);
+
+	/* write zero samples to sink PCM until it is full */
+	while (write(fd_pcm_snk, pcm_zero, sizeof(pcm_zero)) > 0)
+		continue;
+
+	/* verify that the FIFO is not empty */
+	ck_assert_int_eq(ioctl(fd_pcm_snk, FIONREAD, &nread), 0);
+	ck_assert_uint_gt(nread, 0);
+
+	/* drain PCM samples */
+	ck_assert_int_eq(ba_transport_pcm_drain(pcm), 0);
+
+	/* verify that the FIFO has been drained */
+	ck_assert_int_eq(ioctl(fd_pcm_snk, FIONREAD, &nread), 0);
+	ck_assert_uint_eq(nread, 0);
+
+	ba_transport_destroy(t1);
+	ba_transport_destroy(t2);
+	close(fd_pcm_snk);
+	close(fd_pcm_src);
+
+} CK_END_TEST
+
+static void *test_transport_destroy(void *userdata) {
+	struct ba_transport *t = userdata;
+	usleep(10000);
+	ba_transport_destroy(t);
+	return NULL;
+}
+
+CK_START_TEST(test_a2dp_sbc_pcm_drain_and_close) {
+
+	int16_t pcm_zero[90] = { 0 };
+
+	struct ba_transport *t1 = test_transport_new_a2dp(device1,
+			BA_TRANSPORT_PROFILE_A2DP_SOURCE, "/path/sbc", &a2dp_sbc_source,
+			&config_sbc_44100_stereo);
+	struct ba_transport *t2 = test_transport_new_a2dp(device2,
+			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/sbc", &a2dp_sbc_sink,
+			&config_sbc_44100_stereo);
+
+	int fd_pcm_snk = -1;
+	int fd_pcm_src = -1;
+	setup_a2dp_link(t1, t2, 256, &fd_pcm_snk, &fd_pcm_src);
+
+	/* start sink PCM IO thread and make sure it is running */
+	struct ba_transport_pcm *pcm = &t1->a2dp.pcm;
+	ck_assert_int_eq(ba_transport_pcm_start(pcm, a2dp_sbc_enc_thread, "sbc"), 0);
+	ck_assert_int_eq(ba_transport_pcm_state_wait_running(pcm), 0);
+
+	/* write zero samples to sink PCM until it is full */
+	while (write(fd_pcm_snk, pcm_zero, sizeof(pcm_zero)) > 0)
+		continue;
+
+	pthread_t thread;
+	/* schedule transport destruction to kick in during PCM drain */
+	pthread_create(&thread, NULL, test_transport_destroy, ba_transport_ref(t1));
+
+	/* drain PCM samples */
+	ck_assert_int_eq(ba_transport_pcm_drain(pcm), 0);
+
+	pthread_join(thread, NULL);
+
+	ba_transport_destroy(t1);
+	ba_transport_destroy(t2);
+	close(fd_pcm_snk);
+	close(fd_pcm_src);
+
+} CK_END_TEST
+
 CK_START_TEST(test_a2dp_sbc_pcm_drop) {
 
 	int16_t pcm_zero[90] = { 0 };
@@ -719,47 +854,25 @@ CK_START_TEST(test_a2dp_sbc_pcm_drop) {
 	struct ba_transport *t1 = test_transport_new_a2dp(device1,
 			BA_TRANSPORT_PROFILE_A2DP_SOURCE, "/path/sbc", &a2dp_sbc_source,
 			&config_sbc_44100_stereo);
-
 	struct ba_transport *t2 = test_transport_new_a2dp(device2,
 			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/sbc", &a2dp_sbc_sink,
 			&config_sbc_44100_stereo);
 
-	struct ba_transport_thread *th1 = &t1->thread_enc;
-	struct ba_transport_thread *th2 = &t2->thread_dec;
+	int fd_pcm_snk = -1;
+	int fd_pcm_src = -1;
+	setup_a2dp_link(t1, t2, 256, &fd_pcm_snk, &fd_pcm_src);
 
-	int bt_fds[2];
-	ck_assert_int_eq(socketpair(AF_UNIX, SOCK_SEQPACKET | SOCK_NONBLOCK, 0, bt_fds), 0);
-	debug("Created BT socket pair: %d, %d", bt_fds[0], bt_fds[1]);
-	t1->mtu_read = t1->mtu_write = 256;
-	t2->mtu_read = t2->mtu_write = 256;
-	t1->bt_fd = bt_fds[1];
-	t2->bt_fd = bt_fds[0];
-
-	int pcm_snk_fds[2];
-	ck_assert_int_eq(pipe2(pcm_snk_fds, O_NONBLOCK), 0);
-	debug("Created PCM pipe pair: %d, %d", pcm_snk_fds[0], pcm_snk_fds[1]);
-	th1->pcm->fd = pcm_snk_fds[0];
-	th1->pcm->active = true;
-
-	int pcm_src_fds[2];
-	ck_assert_int_eq(pipe2(pcm_src_fds, O_NONBLOCK), 0);
-	debug("Created PCM pipe pair: %d, %d", pcm_src_fds[0], pcm_src_fds[1]);
-	th2->pcm->fd = pcm_src_fds[1];
-	th2->pcm->active = true;
-
-	/* sink PCM */
-	struct ba_transport_pcm *pcm = th1->pcm;
-
-	/* write zero samples to PCM FIFO until it is full */
-	while (write(pcm_snk_fds[1], pcm_zero, sizeof(pcm_zero)) > 0)
+	/* write zero samples to sink PCM until it is full */
+	while (write(fd_pcm_snk, pcm_zero, sizeof(pcm_zero)) > 0)
 		continue;
 
-	/* drop PCM samples before IO thread was started */
+	/* drop sink PCM samples before IO thread was started */
+	struct ba_transport_pcm *pcm = &t1->a2dp.pcm;
 	ck_assert_int_eq(ba_transport_pcm_drop(pcm), 0);
 
 	/* start IO thread and make sure it is running */
-	ck_assert_int_eq(ba_transport_thread_create(th1, a2dp_sbc_enc_thread, "sbc", true), 0);
-	ck_assert_int_eq(ba_transport_thread_state_wait_running(th1), 0);
+	ck_assert_int_eq(ba_transport_pcm_start(pcm, a2dp_sbc_enc_thread, "sbc"), 0);
+	ck_assert_int_eq(ba_transport_pcm_state_wait_running(pcm), 0);
 
 	/* wait for 50 ms - let the thread to run for a while */
 	usleep(50000);
@@ -767,11 +880,11 @@ CK_START_TEST(test_a2dp_sbc_pcm_drop) {
 	uint8_t bt_buffer[1024];
 	/* try to read data from BT socket - it should be empty because
 	 * the PCM has been dropped before the thread was started */
-	ck_assert_int_eq(read(bt_fds[0], bt_buffer, sizeof(bt_buffer)), -1);
+	ck_assert_int_eq(read(t2->bt_fd, bt_buffer, sizeof(bt_buffer)), -1);
 	ck_assert_int_eq(errno, EAGAIN);
 
-	/* write non-zero samples to PCM FIFO and process some of it */
-	while (write(pcm_snk_fds[1], pcm_rand, sizeof(pcm_rand)) > 0)
+	/* write non-zero samples to sink PCM and process some of it */
+	while (write(fd_pcm_snk, pcm_rand, sizeof(pcm_rand)) > 0)
 		continue;
 	usleep(50000);
 
@@ -781,33 +894,33 @@ CK_START_TEST(test_a2dp_sbc_pcm_drop) {
 	/* write a little bit of non-zero samples and wait for processing; the
 	 * number of samples shall be small enough to not produce a single codec
 	 * frame, but enough to fill-in internal buffers */
-	ck_assert_int_gt(write(pcm_snk_fds[1], pcm_rand, sizeof(pcm_rand)), 0);
+	ck_assert_int_gt(write(fd_pcm_snk, pcm_rand, sizeof(pcm_rand)), 0);
 	usleep(10000);
 
 	/* again drop PCM samples */
 	ck_assert_int_eq(ba_transport_pcm_drop(pcm), 0);
 
 	/* flush already processed data */
-	while (read(bt_fds[0], bt_buffer, sizeof(bt_buffer)) > 0)
+	while (read(t2->bt_fd, bt_buffer, sizeof(bt_buffer)) > 0)
 		continue;
 
 	/* After PCM has been dropped, IO thread should not process any more
 	 * non-zero samples. We will check this by writing zero samples and
 	 * checking if decoded data is all silence. */
 
-	ck_assert_int_eq(ba_transport_thread_create(th2, a2dp_sbc_dec_thread, "sbc", true), 0);
-	ck_assert_int_eq(ba_transport_thread_state_wait_running(th2), 0);
+	ck_assert_int_eq(ba_transport_pcm_start(&t2->a2dp.pcm, a2dp_sbc_dec_thread, "sbc"), 0);
+	ck_assert_int_eq(ba_transport_pcm_state_wait_running(&t2->a2dp.pcm), 0);
 
-	/* write some zero samples to PCM FIFO and process them */
+	/* write some zero samples to sink PCM and process them */
 	for (size_t i = 0; i < 100; i++)
-		if (write(pcm_snk_fds[1], pcm_zero, sizeof(pcm_zero)) <= 0)
+		if (write(fd_pcm_snk, pcm_zero, sizeof(pcm_zero)) <= 0)
 			break;
 	usleep(250000);
 
 	ssize_t rv;
 	int16_t pcm_buffer[1024];
 	/* read decoded data and check if it is all silence */
-	while ((rv = read(pcm_src_fds[0], pcm_buffer, sizeof(pcm_buffer))) > 0) {
+	while ((rv = read(fd_pcm_src, pcm_buffer, sizeof(pcm_buffer))) > 0) {
 		const size_t samples = rv / sizeof(pcm_buffer[0]);
 		for (size_t i = 0; i < samples; i++)
 			ck_assert_int_eq(pcm_buffer[i], 0);
@@ -815,9 +928,8 @@ CK_START_TEST(test_a2dp_sbc_pcm_drop) {
 
 	ba_transport_destroy(t1);
 	ba_transport_destroy(t2);
-	close(pcm_snk_fds[0]);
-	close(pcm_src_fds[1]);
-	close(bt_fds[0]);
+	close(fd_pcm_snk);
+	close(fd_pcm_src);
 
 } CK_END_TEST
 
@@ -833,14 +945,17 @@ CK_START_TEST(test_a2dp_mp3) {
 			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/mp3", &a2dp_mpeg_sink,
 			&config_mp3_44100_stereo);
 
+	struct ba_transport_pcm *t1_pcm = &t1->a2dp.pcm;
+	struct ba_transport_pcm *t2_pcm = &t2->a2dp.pcm;
+
 	if (aging_duration) {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 1024;
-		test_io(t1, t2, a2dp_mp3_enc_thread, a2dp_mpeg_dec_thread, 4 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_mp3_enc_thread, a2dp_mpeg_dec_thread, 4 * 1024);
 	}
 	else {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 512;
-		test_io(t1, t2, a2dp_mp3_enc_thread, test_io_thread_dump_bt, 3 * 1024);
-		test_io(t1, t2, test_io_thread_dump_pcm, a2dp_mpeg_dec_thread, 3 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_mp3_enc_thread, test_io_thread_dump_bt, 3 * 1024);
+		test_io(t1_pcm, t2_pcm, test_io_thread_dump_pcm, a2dp_mpeg_dec_thread, 3 * 1024);
 	}
 
 	ba_transport_destroy(t1);
@@ -863,14 +978,17 @@ CK_START_TEST(test_a2dp_aac) {
 			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/aac", &a2dp_aac_sink,
 			&config_aac_44100_stereo);
 
+	struct ba_transport_pcm *t1_pcm = &t1->a2dp.pcm;
+	struct ba_transport_pcm *t2_pcm = &t2->a2dp.pcm;
+
 	if (aging_duration) {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 450;
-		test_io(t1, t2, a2dp_aac_enc_thread, a2dp_aac_dec_thread, 4 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_aac_enc_thread, a2dp_aac_dec_thread, 4 * 1024);
 	}
 	else {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 190;
-		test_io(t1, t2, a2dp_aac_enc_thread, test_io_thread_dump_bt, 2 * 1024);
-		test_io(t1, t2, test_io_thread_dump_pcm, a2dp_aac_dec_thread, 2 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_aac_enc_thread, test_io_thread_dump_bt, 2 * 1024);
+		test_io(t1_pcm, t2_pcm, test_io_thread_dump_pcm, a2dp_aac_dec_thread, 2 * 1024);
 	}
 
 	ba_transport_destroy(t1);
@@ -889,17 +1007,20 @@ CK_START_TEST(test_a2dp_aptx) {
 			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/aptx", &a2dp_aptx_sink,
 			&config_aptx_44100_stereo);
 
+	struct ba_transport_pcm *t1_pcm = &t1->a2dp.pcm;
+	struct ba_transport_pcm *t2_pcm = &t2->a2dp.pcm;
+
 	if (aging_duration) {
 #if HAVE_APTX_DECODE
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 400;
-		test_io(t1, t2, a2dp_aptx_enc_thread, a2dp_aptx_dec_thread, 4 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_aptx_enc_thread, a2dp_aptx_dec_thread, 4 * 1024);
 #endif
 	}
 	else {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 40;
-		test_io(t1, t2, a2dp_aptx_enc_thread, test_io_thread_dump_bt, 2 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_aptx_enc_thread, test_io_thread_dump_bt, 2 * 1024);
 #if HAVE_APTX_DECODE
-		test_io(t1, t2, test_io_thread_dump_pcm, a2dp_aptx_dec_thread, 2 * 1024);
+		test_io(t1_pcm, t2_pcm, test_io_thread_dump_pcm, a2dp_aptx_dec_thread, 2 * 1024);
 #endif
 	};
 
@@ -919,17 +1040,20 @@ CK_START_TEST(test_a2dp_aptx_hd) {
 			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/aptxhd", &a2dp_aptx_hd_sink,
 			&config_aptx_hd_44100_stereo);
 
+	struct ba_transport_pcm *t1_pcm = &t1->a2dp.pcm;
+	struct ba_transport_pcm *t2_pcm = &t2->a2dp.pcm;
+
 	if (aging_duration) {
 #if HAVE_APTX_HD_DECODE
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 600;
-		test_io(t1, t2, a2dp_aptx_hd_enc_thread, a2dp_aptx_hd_dec_thread, 4 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_aptx_hd_enc_thread, a2dp_aptx_hd_dec_thread, 4 * 1024);
 #endif
 	}
 	else {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 60;
-		test_io(t1, t2, a2dp_aptx_hd_enc_thread, test_io_thread_dump_bt, 2 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_aptx_hd_enc_thread, test_io_thread_dump_bt, 2 * 1024);
 #if HAVE_APTX_HD_DECODE
-		test_io(t1, t2, test_io_thread_dump_pcm, a2dp_aptx_hd_dec_thread, 2 * 1024);
+		test_io(t1_pcm, t2_pcm, test_io_thread_dump_pcm, a2dp_aptx_hd_dec_thread, 2 * 1024);
 #endif
 	};
 
@@ -949,14 +1073,17 @@ CK_START_TEST(test_a2dp_faststream_music) {
 			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/faststream", &a2dp_faststream_sink,
 			&config_faststream_44100_16000);
 
+	struct ba_transport_pcm *t1_pcm = &t1->a2dp.pcm;
+	struct ba_transport_pcm *t2_pcm = &t2->a2dp.pcm;
+
 	if (aging_duration) {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 72 * 3;
-		test_io(t1, t2, a2dp_faststream_enc_thread, a2dp_faststream_dec_thread, 4 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_faststream_enc_thread, a2dp_faststream_dec_thread, 4 * 1024);
 	}
 	else {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 72 * 3;
-		test_io(t1, t2, a2dp_faststream_enc_thread, test_io_thread_dump_bt, 2 * 1024);
-		test_io(t1, t2, test_io_thread_dump_pcm, a2dp_faststream_dec_thread, 2 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_faststream_enc_thread, test_io_thread_dump_bt, 2 * 1024);
+		test_io(t1_pcm, t2_pcm, test_io_thread_dump_pcm, a2dp_faststream_dec_thread, 2 * 1024);
 	}
 
 	ba_transport_destroy(t1);
@@ -975,14 +1102,17 @@ CK_START_TEST(test_a2dp_faststream_voice) {
 			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/faststream", &a2dp_faststream_sink,
 			&config_faststream_44100_16000);
 
+	struct ba_transport_pcm *t1_pcm_bc = &t1->a2dp.pcm_bc;
+	struct ba_transport_pcm *t2_pcm_bc = &t2->a2dp.pcm_bc;
+
 	if (aging_duration) {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 72 * 3;
-		test_io(t2, t1, a2dp_faststream_enc_thread, a2dp_faststream_dec_thread, 4 * 1024);
+		test_io(t2_pcm_bc, t1_pcm_bc, a2dp_faststream_enc_thread, a2dp_faststream_dec_thread, 4 * 1024);
 	}
 	else {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 72 * 3;
-		test_io(t2, t1, a2dp_faststream_enc_thread, test_io_thread_dump_bt, 2 * 1024);
-		test_io(t2, t1, test_io_thread_dump_pcm, a2dp_faststream_dec_thread, 2 * 1024);
+		test_io(t2_pcm_bc, t1_pcm_bc, a2dp_faststream_enc_thread, test_io_thread_dump_bt, 2 * 1024);
+		test_io(t2_pcm_bc, t1_pcm_bc, test_io_thread_dump_pcm, a2dp_faststream_dec_thread, 2 * 1024);
 	}
 
 	ba_transport_destroy(t1);
@@ -1001,16 +1131,19 @@ CK_START_TEST(test_a2dp_lc3plus) {
 			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/lc3plus", &a2dp_lc3plus_sink,
 			&config_lc3plus_48000_stereo);
 
+	struct ba_transport_pcm *t1_pcm = &t1->a2dp.pcm;
+	struct ba_transport_pcm *t2_pcm = &t2->a2dp.pcm;
+
 	if (aging_duration) {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write =
 			RTP_HEADER_LEN + sizeof(rtp_media_header_t) + 300;
-		test_io(t1, t2, a2dp_lc3plus_enc_thread, a2dp_lc3plus_dec_thread, 4 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_lc3plus_enc_thread, a2dp_lc3plus_dec_thread, 4 * 1024);
 	}
 	else {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write =
 			RTP_HEADER_LEN + sizeof(rtp_media_header_t) + 300;
-		test_io(t1, t2, a2dp_lc3plus_enc_thread, test_io_thread_dump_bt, 2 * 1024);
-		test_io(t1, t2, test_io_thread_dump_pcm, a2dp_lc3plus_dec_thread, 2 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_lc3plus_enc_thread, test_io_thread_dump_bt, 2 * 1024);
+		test_io(t1_pcm, t2_pcm, test_io_thread_dump_pcm, a2dp_lc3plus_dec_thread, 2 * 1024);
 	}
 
 	ba_transport_destroy(t1);
@@ -1032,19 +1165,22 @@ CK_START_TEST(test_a2dp_ldac) {
 			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/ldac", &a2dp_ldac_sink,
 			&config_ldac_44100_stereo);
 
+	struct ba_transport_pcm *t1_pcm = &t1->a2dp.pcm;
+	struct ba_transport_pcm *t2_pcm = &t2->a2dp.pcm;
+
 	if (aging_duration) {
 #if HAVE_LDAC_DECODE
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write =
 			RTP_HEADER_LEN + sizeof(rtp_media_header_t) + 990 + 6;
-		test_io(t1, t2, a2dp_ldac_enc_thread, a2dp_ldac_dec_thread, 4 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_ldac_enc_thread, a2dp_ldac_dec_thread, 4 * 1024);
 #endif
 	}
 	else {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write =
 			RTP_HEADER_LEN + sizeof(rtp_media_header_t) + 660 + 6;
-		test_io(t1, t2, a2dp_ldac_enc_thread, test_io_thread_dump_bt, 2 * 1024);
+		test_io(t1_pcm, t2_pcm, a2dp_ldac_enc_thread, test_io_thread_dump_bt, 2 * 1024);
 #if HAVE_LDAC_DECODE
-		test_io(t1, t2, test_io_thread_dump_pcm, a2dp_ldac_dec_thread, 2 * 1024);
+		test_io(t1_pcm, t2_pcm, test_io_thread_dump_pcm, a2dp_ldac_dec_thread, 2 * 1024);
 #endif
 	}
 
@@ -1061,9 +1197,12 @@ CK_START_TEST(test_sco_cvsd) {
 	struct ba_transport *t2 = test_transport_new_sco(device2,
 			BA_TRANSPORT_PROFILE_HSP_AG, "/path/sco/cvsd");
 
+	struct ba_transport_pcm *t1_pcm = &t1->sco.pcm_spk;
+	struct ba_transport_pcm *t2_pcm = &t2->sco.pcm_spk;
+
 	t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 48;
-	test_io(t1, t2, sco_enc_thread, test_io_thread_dump_bt, 600);
-	test_io(t1, t2, test_io_thread_dump_pcm, sco_dec_thread, 600);
+	test_io(t1_pcm, t2_pcm, sco_enc_thread, test_io_thread_dump_bt, 600);
+	test_io(t1_pcm, t2_pcm, test_io_thread_dump_pcm, sco_dec_thread, 600);
 
 	ba_transport_destroy(t1);
 	ba_transport_destroy(t2);
@@ -1075,6 +1214,7 @@ CK_START_TEST(test_sco_msbc) {
 
 	adapter->hci.features[2] = LMP_TRSP_SCO;
 	adapter->hci.features[3] = LMP_ESCO;
+	config.io_thread_rt_priority = 10;
 
 	struct ba_transport *t1 = test_transport_new_sco(device1,
 			BA_TRANSPORT_PROFILE_HFP_AG, "/path/sco/msbc");
@@ -1083,9 +1223,38 @@ CK_START_TEST(test_sco_msbc) {
 			BA_TRANSPORT_PROFILE_HFP_AG, "/path/sco/msbc");
 	ba_transport_set_codec(t2, HFP_CODEC_MSBC);
 
+	struct ba_transport_pcm *t1_pcm = &t1->sco.pcm_spk;
+	struct ba_transport_pcm *t2_pcm = &t2->sco.pcm_spk;
+
 	t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 24;
-	test_io(t1, t2, sco_enc_thread, test_io_thread_dump_bt, 600);
-	test_io(t1, t2, test_io_thread_dump_pcm, sco_dec_thread, 600);
+	test_io(t1_pcm, t2_pcm, sco_enc_thread, test_io_thread_dump_bt, 600);
+	test_io(t1_pcm, t2_pcm, test_io_thread_dump_pcm, sco_dec_thread, 600);
+
+	ba_transport_destroy(t1);
+	ba_transport_destroy(t2);
+
+} CK_END_TEST
+#endif
+
+#if ENABLE_LC3_SWB
+CK_START_TEST(test_sco_lc3_swb) {
+
+	adapter->hci.features[2] = LMP_TRSP_SCO;
+	adapter->hci.features[3] = LMP_ESCO;
+
+	struct ba_transport *t1 = test_transport_new_sco(device1,
+			BA_TRANSPORT_PROFILE_HFP_AG, "/path/sco/lc3swb");
+	ba_transport_set_codec(t1, HFP_CODEC_LC3_SWB);
+	struct ba_transport *t2 = test_transport_new_sco(device2,
+			BA_TRANSPORT_PROFILE_HFP_AG, "/path/sco/lc3swb");
+	ba_transport_set_codec(t2, HFP_CODEC_LC3_SWB);
+
+	struct ba_transport_pcm *t1_pcm = &t1->sco.pcm_spk;
+	struct ba_transport_pcm *t2_pcm = &t2->sco.pcm_spk;
+
+	t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 24;
+	test_io(t1_pcm, t2_pcm, sco_enc_thread, test_io_thread_dump_bt, 600);
+	test_io(t1_pcm, t2_pcm, test_io_thread_dump_pcm, sco_dec_thread, 600);
 
 	ba_transport_destroy(t1);
 	ba_transport_destroy(t2);
@@ -1105,6 +1274,8 @@ int main(int argc, char *argv[]) {
 	} codecs[] = {
 		{ a2dp_codecs_codec_id_to_string(A2DP_CODEC_SBC), test_a2dp_sbc },
 		{ a2dp_codecs_codec_id_to_string(A2DP_CODEC_SBC), test_a2dp_sbc_invalid_config },
+		{ a2dp_codecs_codec_id_to_string(A2DP_CODEC_SBC), test_a2dp_sbc_pcm_drain },
+		{ a2dp_codecs_codec_id_to_string(A2DP_CODEC_SBC), test_a2dp_sbc_pcm_drain_and_close },
 		{ a2dp_codecs_codec_id_to_string(A2DP_CODEC_SBC), test_a2dp_sbc_pcm_drop },
 #if ENABLE_MP3LAME
 		{ a2dp_codecs_codec_id_to_string(A2DP_CODEC_MPEG12), test_a2dp_mp3 },
@@ -1131,6 +1302,9 @@ int main(int argc, char *argv[]) {
 		{ hfp_codec_id_to_string(HFP_CODEC_CVSD), test_sco_cvsd },
 #if ENABLE_MSBC
 		{ hfp_codec_id_to_string(HFP_CODEC_MSBC), test_sco_msbc },
+#endif
+#if ENABLE_LC3_SWB
+		{ hfp_codec_id_to_string(HFP_CODEC_LC3_SWB), test_sco_lc3_swb },
 #endif
 	};
 
@@ -1224,7 +1398,7 @@ int main(int argc, char *argv[]) {
 
 	}
 
-	bluealsa_config_init();
+	ba_config_init();
 
 	bdaddr_t addr1 = {{ 1, 2, 3, 4, 5, 6 }};
 	bdaddr_t addr2 = {{ 1, 2, 3, 7, 8, 9 }};
@@ -1248,7 +1422,11 @@ int main(int argc, char *argv[]) {
 
 	srunner_run_all(sr, CK_ENV);
 	int nf = srunner_ntests_failed(sr);
+
 	srunner_free(sr);
+	ba_device_unref(device1);
+	ba_device_unref(device2);
+	ba_adapter_unref(adapter);
 
 	return nf == 0 ? 0 : 1;
 }
